"""Functions for generating stock & flow diagrams.

Diagrams are graphviz dot diagrams, and notably don't follow the
exact format typically used with SFDs, namely sources and sinks
aren't explicitly represented.
"""

import os

import matplotlib.pyplot as plt
import xarray as xr
from graphviz import Digraph

import reno
from reno.components import Flow, Stock, TimeRef, TrackedReference, Variable

# Theming constants, set for lightmode by default.
DARK_MODE = False
"""When ``True``, set graph node_attr to use colors more suited for dark backgrounds."""
SUBGRAPH_COLORS = ["#BBDDFF", "#DDBBFF"]
"""Container background colors for nested models/submodels, each subsquent
level will go one color farther into this array.

A decent option for dark mode: ``["#334455", "#443355"]``
"""
NODE_COLOR = "lightgreen"
"""Background color to use for variable nodes.

A decent option for dark mode: ``"darkgreen"``
"""


def set_dark_mode(dark: bool = False):
    """Set themeing constants for the diagrams appropriately for
    dark/light theme.

    Note that this will also change matplotlib style so that spark plots
    match. Set the ``DARK_MODE``, ``SUBGRAPH_COLORS``, and ``NODE_COLOR``
    constants manually if this is undesired.

    Args:
        dark (bool): Pass ``False`` for light theme, ``True`` for dark theme.
    """

    global DARK_MODE, SUBGRAPH_COLORS, NODE_COLOR
    DARK_MODE = dark
    if not dark:
        plt.style.use("default")
        SUBGRAPH_COLORS = ["#BBDDFF", "#DDBBFF"]
        NODE_COLOR = "lightgreen"
    else:
        plt.style.use("dark_background")
        SUBGRAPH_COLORS = ["#334455", "#443355"]
        NODE_COLOR = "darkgreen"


def stock_flow_diagram(
    model,
    show_vars: bool = True,
    exclude_var_names: list[str] = None,
    sparklines: bool = False,
    sparkdensities: bool = False,
    sparkall: bool = False,
    g: Digraph = None,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
    lr: bool = False,
    _level: int = 0,
) -> tuple[Digraph, list[tuple[str, str]]]:
    """Generate a graphviz dot graph for all the stocks and flows of the passed model,
    optionally including sparklines if a simulation has been run.

    Args:
        model (reno.model.Model): The model to collect stocks/flows/variables from.
        show_vars (bool): Whether to render variables in the diagram, or just stocks and flows
            (for very complex models, hiding variables can make it a bit easier to visually
            parse.)
        exclude_var_names (list[str]): Specific variables to hide in the diagram, can be used
            with ``show_vars=True`` to just show specific variables of interest.
        sparklines (bool): Draw mini graphs to the right of each stock showing plotting their
            values through time. This assumes the model has either been run, or traces are passed
            in manually with the ``traces`` argument.
        sparkdensities (bool): Draw mini density plots/histograms next to any variables that
            sample from distributions. This assumes the model has either been run, or traces
            are passed in manually with the ``traces`` argument.
        g (Digraph): Graphviz Digraph instance to render the nodes/edges on. Mostly only for
            internal use for drawing subgraphs for submodels.
        traces (list[xr.Dataset]): A list of traces or model run datasets to use for drawing
            spark plots. Each dataset will be rendered in a different color.
        universe (list[TrackedReference]): Limit rendered nodes to only those listed here, this
            includes all of stocks/flows/variables. (This acts as an initial filter, ``show_vars``
            and ``exclude_var_names`` still applies after this.)
        lr (bool): By default the graphviz plot tries to orient top-down. Specify ``True`` to
            try to orient it left-right.

    Returns:
        The populated Digraph instance (Jupyter can natively render this in a cell output.) and
        a list of tuples of reference connections that weren't added (usually because one of the
        references was "out of scope" part of a parent model not rendered in this one. This is
        primarily for internal use for correctly recursively plotting submodels in subgraphs.
    """
    if g is None:
        rankdir = "TB" if not lr else "LR"
        if DARK_MODE:
            g = Digraph(
                name=model.name,
                graph_attr=dict(rankdir=rankdir, bgcolor="#181818"),
                node_attr=dict(
                    fontcolor="#e6e6e6",
                    style="filled",
                    color="#e6e6e6",
                    fillcolor="#333333",
                ),
                edge_attr=dict(color="#e6e6e6", fontcolor="#e6e6e6"),
            )
        else:
            g = Digraph(name=model.name, graph_attr=dict(rankdir=rankdir))
    if exclude_var_names is None:
        exclude_var_names = []

    g, out_of_scope_stock_connections = add_stocks(
        g, model.stocks, sparklines=sparklines, traces=traces, universe=universe
    )
    out_of_scope_var_connections = []
    if show_vars:
        variables = filter_variables(model.vars, exclude_var_names)
        g, out_of_scope_var_connections = add_vars(
            g,
            variables,
            exclude_var_names,
            sparkdensities,
            sparkall,
            traces,
            universe=universe,
        )
    else:
        variables = []
    g, out_of_scope_flow_connections = add_flows(
        g, model.flows, variables, sparkall, traces, universe=universe
    )

    missing_connections = (
        out_of_scope_var_connections
        + out_of_scope_flow_connections
        + out_of_scope_stock_connections
    )

    # recursively render any submodels.
    for submodel in model.models:
        with g.subgraph(name=submodel.name) as g_s:
            g_s.attr(style="filled")
            g_s.attr(color=SUBGRAPH_COLORS[_level])
            g_s.attr(cluster="true")
            g_s.attr(label=submodel.label)
            g_s.attr(fontcolor="#888888")
            m = getattr(model, submodel.name)
            unused_graph, remaining_connections = stock_flow_diagram(
                m,
                show_vars=show_vars,
                exclude_var_names=exclude_var_names,
                sparklines=sparklines,
                sparkdensities=sparkdensities,
                sparkall=sparkall,
                g=g_s,
                traces=traces,
                universe=universe,
                lr=lr,
                _level=_level + 1,
            )
            for src, dst in remaining_connections:
                # check if we are at the correct level model to draw this,
                # or do we pass it on up?
                if src.model == model or dst.model == model:
                    draw_appropriate_edge(g, src, dst)
                else:
                    missing_connections.append((src, dst))

    return g, missing_connections


def add_stock_io_edge(g: Digraph, stock: Stock, flow: Flow):
    """Edge with attributes for heavily highlighting in flows and
    outflows from stocks."""
    if flow in stock.in_flows:
        name1 = flow.qual_name()
        name2 = stock.qual_name()
    elif flow in stock.out_flows:
        name1 = stock.qual_name()
        name2 = flow.qual_name()
    else:
        raise Exception(
            f"Could not draw edge for flow {flow.qual_name()} and stock {stock.qual_name()}, stock does not use flow."
        )
    g.edge(name1, name2, style="bold", weight="50")


def add_stock_limit_edge(g: Digraph, ref, stock: Stock):
    """De-emphasize stock variable/flow connections if they're only
    used in min/max constraints."""
    g.edge(
        ref.qual_name(),
        stock.qual_name(),
        style="dotted",
        arrowsize=".5",
    )


def add_to_flow_edge(g: Digraph, ref, flow: Flow):
    """Non-stock-inflow/outflow-related edges that point to flows
    are slightly de-emphasized."""
    style = "dotted" if isinstance(ref, Variable) else "dashed"
    constraint = "false" if isinstance(ref, Stock) else "true"
    # TODO: constraint should maybe take same model scope into account?
    # (stock to out flows edges already handled, any others
    # shouldn't affect rank)
    g.edge(
        ref.qual_name(),
        flow.qual_name(),
        style=style,
        arrowsize=".5",
        constraint=constraint,
    )


def add_to_var_edge(g: Digraph, ref, variable: Variable):
    """Edges pointing at variables are de-emphasized (usually these are
    auxiliary variables - variables calculated based on other variables.)"""
    g.edge(
        ref.qual_name(),
        variable.qual_name(),
        style="dotted",
        arrowsize=".5",
    )


# this is mostly only important for determining how to draw edges between refs
# in different scopes
# TODO: pretty sure there's still a mistake in this, will sometimes still try to
# include out of scope variable and just gets added as default ellipse.
def draw_appropriate_edge(g: Digraph, src: TrackedReference, dst: TrackedReference):
    """Switch between edge types correctly based on the types of src/dst nodes."""
    if (
        isinstance(src, Flow)
        and isinstance(dst, Stock)
        and (src in dst.in_flows or src in dst.out_flows)
    ):
        add_stock_io_edge(g, dst, src)
    elif (
        isinstance(src, Stock)
        and isinstance(dst, Flow)
        and (dst in src.in_flows or dst in src.out_flows)
    ):
        add_stock_io_edge(g, src, dst)
    elif isinstance(dst, Stock) and (src in dst.min_refs() or src in dst.max_refs()):
        add_stock_limit_edge(g, src, dst)
    elif isinstance(dst, Flow):
        add_to_flow_edge(g, src, dst)
    elif isinstance(dst, Variable):
        add_to_var_edge(g, src, dst)
    else:
        raise Exception(
            f"What type of nodes are you asking me to connect??? {src} -> {dst}?"
        )


def filter_variables(
    all_variables: list[Variable], exclude_names: list[str]
) -> list[Variable]:
    """Take a list of variables and a list of names of variables to remove from the list.

    Returns the list of variables that do NOT have a name in the exclude_names list.
    """
    filtered = [
        variable for variable in all_variables if variable.name not in exclude_names
    ]
    return filtered


def add_stocks(
    g: Digraph,
    stocks: list[Stock],
    sparklines: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
):
    """Add nodes and edges for all passed stocks to the passed graph.

    Args:
        sparklines (bool): include small little line graph of each stock in the
            diagram. This only works if a simulation has been run.
    """

    render_in_parent_scope = []

    for stock in stocks:
        if universe is not None and stock not in universe:
            continue

        # add a node for each stock
        if not sparklines:
            g.node(
                name=stock.qual_name(),
                label=stock.label,
                shape="rect",
                group=stock.group,
            )
        else:
            # to force the sparkline graph image node to render right next to
            # its stock node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same"}) as c:
                c.node(
                    name=stock.qual_name(),
                    label=stock.label,
                    shape="rect",
                    group=stock.group,
                )

                # generate the sparkline graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))
                    # use the bayes one if available
                    if stock.model.trace is not None or traces is not None:
                        if traces is None:
                            traces = [stock.model.trace.prior]
                            if "posterior" in stock.model.trace:
                                traces.append(stock.model.trace.posterior)

                        reno.viz.compare_seq(
                            stock.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    else:
                        # TODO: can probably still use compare_seq for regular
                        # reno run plot? (since traces can be xrDataset)
                        stock.plot(ax)
                    ax.xaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{stock.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{stock.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{stock.qual_name()}.png",
                    shape="none",
                    group=stock.group,
                )
                c.edge(
                    stock.qual_name(),
                    f"{stock.qual_name()}_fig",
                    constraint="true",
                    weight="20",
                    dir="none",
                )

        # look at each stock's inflows/outflows and add the appropriate edges
        # (this works even if the flow "nodes" haven't been added yet.)
        handled_refs = []
        for flow in stock.in_flows:
            if universe is not None and flow not in universe:
                continue
            if reno.utils.is_ref_in_parent_scope(flow, stock):
                # add refs themselves not qualname, need to access model
                render_in_parent_scope.append((flow, stock))
                continue
            add_stock_io_edge(g, stock, flow)
            handled_refs.append(flow.qual_name())
        for flow in stock.out_flows:
            if universe is not None and flow not in universe:
                continue
            if reno.utils.is_ref_in_parent_scope(flow, stock):
                # add refs themselves not qualname, need to access model
                render_in_parent_scope.append((flow, stock))
                continue
            add_stock_io_edge(g, stock, flow)
            handled_refs.append(flow.qual_name())
        remaining = []
        remaining.extend(stock.min_refs())
        remaining.extend(stock.max_refs())
        for ref in remaining:
            if ref not in handled_refs:
                if universe is not None and flow not in universe:
                    continue
                if reno.utils.is_ref_in_parent_scope(ref, stock):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, stock))
                    continue
                add_stock_limit_edge(g, ref, stock)
    return g, render_in_parent_scope


def add_vars(
    g: Digraph,
    variables: list[Variable],
    exclude_vars: list[str],
    sparkdensities: bool = False,
    sparkall: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
):
    """Add variables and edges between variables to the passed graphviz graph."""
    rendered_edges = []
    render_in_parent_scope = []
    for variable in variables:
        if universe is not None and variable not in universe:
            continue
        if not sparkdensities or (
            sparkdensities
            and (variable.qual_name() not in variable.model.trace_RVs and not sparkall)
        ):
            g.node(
                name=variable.qual_name(),
                label=variable.label,
                style="rounded,filled",
                fillcolor=NODE_COLOR,
                shape="rect",
                fontsize="10pt",
                height=".2",
                group=variable.group,
            )
        else:
            # to force the sparkdensity graph image node to render right next to
            # its variable node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same"}) as c:
                c.node(
                    name=variable.qual_name(),
                    label=variable.label,
                    style="rounded,filled",
                    fillcolor=NODE_COLOR,
                    shape="rect",
                    fontsize="10pt",
                    height=".2",
                    group=variable.group,
                )

                # generate sparkdensity graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))

                    if traces is None:
                        traces = [variable.model.trace.prior]
                        if "posterior" in variable.model.trace:
                            traces.append(variable.model.trace.posterior)

                    reno.viz.compare_posterior(
                        variable.qual_name(), traces, ax=ax, legend=False, title=""
                    )

                    ax.yaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{variable.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{variable.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{variable.qual_name()}.png",
                    shape="none",
                    group=variable.group,
                )
                c.edge(
                    variable.qual_name(),
                    f"{variable.qual_name()}_fig",
                    constraint="true",
                    weight="20",
                    dir="none",
                )

        # add any edges between variables
        # -- unclear if correct --
        for ref in variable.seek_refs():
            if universe is not None and ref not in universe:
                continue
            if (
                ref.name not in exclude_vars
                and (ref.qual_name(), variable.qual_name()) not in rendered_edges
            ):
                # don't render connections to things in parent model - parent
                # model should handle
                if reno.utils.is_ref_in_parent_scope(ref, variable):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, variable))
                    continue

                add_to_var_edge(g, ref, variable)
                rendered_edges.append((ref.qual_name(), variable.qual_name()))
        # -- /unclear if correct --
    return g, render_in_parent_scope


def add_flows(
    g: Digraph,
    flows: list[Flow],
    variables: list[Variable],
    sparkall: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
):
    """Add flows and edges from variables to the passed graphviz graph."""
    rendered_edges = []

    render_in_parent_scope = []

    for flow in flows:
        if universe is not None and flow not in universe:
            continue
        if not sparkall:
            g.node(
                name=flow.qual_name(), label=flow.name, shape="plain", group=flow.group
            )
        else:
            # to force the sparkline graph image node to render right next to
            # its flow node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same"}) as c:
                c.node(
                    name=flow.qual_name(),
                    label=flow.name,
                    shape="plain",
                    group=flow.group,
                )

                # generate the sparkline graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))
                    # use the bayes one if available
                    if flow.model.trace is not None or traces is not None:
                        if traces is None:
                            traces = [flow.model.trace.prior]
                            if "posterior" in flow.model.trace:
                                traces.append(flow.model.trace.posterior)

                        reno.viz.compare_seq(
                            flow.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    else:
                        # TODO: can probably still use compare_seq for regular
                        # reno run plot? (since traces can be xrDataset)
                        # TODO: TODO: TODO: flow.plot doesn't exist?!
                        # flow.plot(ax)
                        if traces is None:
                            ds = flow.model.dataset()
                            traces = [ds]
                        reno.viz.compare_seq(
                            flow.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    ax.xaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{flow.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{flow.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{flow.qual_name()}.png",
                    shape="none",
                    group=flow.group,
                )
                c.edge(
                    flow.qual_name(),
                    f"{flow.qual_name()}_fig",
                    constraint="true",
                    weight="20",
                    dir="none",
                )

        for ref in flow.seek_refs():
            if universe is not None and ref not in universe:
                continue
            if (
                (
                    ref in variables
                    # TODO: a function called before this that removes variables
                    # of name x
                    and (ref.qual_name(), flow.qual_name()) not in rendered_edges
                )
                # TODO: ...what?
                or (
                    ref not in variables
                    and not isinstance(
                        ref, (TimeRef, Variable)
                    )  # TODO: Variable wasn't here before, with this here means we won't be able to show connections between submodels?
                    and (ref.qual_name(), flow.qual_name()) not in rendered_edges
                    # don't re-render a line from a stock to a flow if the flow
                    # is explicitly an outflow of the stock
                    and not (isinstance(ref, Stock) and flow in ref.out_flows)
                )
            ):
                # don't render connections to things in parent model - parent
                # model should handle
                if reno.utils.is_ref_in_parent_scope(ref, flow):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, flow))
                    continue

                add_to_flow_edge(g, ref, flow)
                rendered_edges.append((ref.qual_name(), flow.qual_name()))

    return g, render_in_parent_scope



"""Tests for the math operations."""

import numpy as np

from reno import model, ops
from reno.components import Piecewise, PostMeasurement, Scalar, TimeRef, Variable


def test_sum_on_matrix():
    """Running ops.sum on a matrix should give you a row-wise sum."""
    v = Variable()
    v.value = np.array([[0, 1, 2], [1, 2, 3]])

    assert (ops.sum(v).eval(3) == np.array([[3, 6]])).all()


def test_sum_on_vector():
    """Running ops.sum on a vector (static variable) should give you the row-wise
    "sum" which is just the value times the number of timesteps."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (ops.sum(v).eval(3) == np.array([[8, 12]])).all()


def test_sum_on_matrix_start_stop():
    """Running ops.sum on a matrix should give you a row-wise sum, correctly
    accounting for specified range."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v[2:].sum().eval(5) == np.array([[9, 12]])).all()
    assert (v[:2].sum().eval(5) == np.array([[1, 3]])).all()
    assert (v[1:3].sum().eval(5) == np.array([[3, 5]])).all()


def test_sum_on_vector_start_stop():
    """Running ops.sum on a vector (static variable) should give you the row-wise
    "sum" which is just the value times the number of timesteps. Correctly
    accounting for specified range."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (v[2:].sum().eval(4) == np.array([[6, 9]])).all()
    assert (v[:2].sum().eval(4) == np.array([[4, 6]])).all()
    assert (v[1:3].sum().eval(4) == np.array([[4, 6]])).all()


def test_static_value_sum():
    """By default, the sum of a static value will just be itself, unless you access
    it as a slice.
    NOTE: changed this behavior by having sum automatically insert slice if static detected.
    """
    v = Variable(Scalar(5))

    assert v.sum().eval(4) == np.array([25])
    assert v[:].sum().eval(4) == np.array([25])


def test_series_max():
    """Both API forms should return the correct series max."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v.series_max().eval(4) == ops.series_max(v).eval(4)).all()


def test_slice_on_matrix():
    """Slice bounds should work properly on a matrix."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v[:2].eval(5) == np.array([[0, 1], [1, 2]])).all()
    assert v[:2].eval(5).shape == np.array([[0, 1], [1, 2]]).shape
    assert (v[0:2].eval(5) == np.array([[0, 1], [1, 2]])).all()
    assert v[0:2].eval(5).shape == np.array([[0, 1], [1, 2]]).shape
    assert (v[3:].eval(5) == np.array([[3, 4], [4, 5]])).all()
    assert v[3:].eval(5).shape == np.array([[3, 4], [4, 5]]).shape


def test_slice_on_vector():
    """Slice bounds should work properly on a matrix."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (v[:2].eval(3) == np.array([[2, 2], [3, 3]])).all()
    assert v[:2].eval(3).shape == np.array([[2, 2], [3, 3]]).shape
    assert (v[0:2].eval(3) == np.array([[2, 2], [3, 3]])).all()
    assert v[0:2].eval(3).shape == np.array([[2, 2], [3, 3]]).shape
    assert (v[3:].eval(5) == np.array([[2, 2, 2], [3, 3, 3]])).all()
    assert v[3:].eval(5).shape == np.array([[2, 2, 2], [3, 3, 3]]).shape


def test_slice_on_scalar():
    """Slicing a scalar should result in a 2d array. (Otherwise
    series ops won't work correctly on the expanded slice.)"""
    v = Variable(Scalar(5))
    assert (v[:].eval(4) == np.array([[5, 5, 5, 5, 5]])).all()
    assert v[:].eval(4).shape == np.array([[5, 5, 5, 5, 5]]).shape


def test_slice_staticness_static_slice():
    """A slice with defined bounds (or static bounds) should be considered static."""
    v = Variable(None)
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert v[1:3].is_static()

    a = Scalar(1)
    b = Scalar(3)

    assert v[a:b].is_static()

    av = Variable(a)
    bv = Variable(b)

    assert v[av:bv].is_static()


def test_slice_staticness_none_endpoint():
    """A slice should not be considered static if it's endpoint is None or t."""
    v = Variable(None)
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert not v[:].is_static()
    assert not v[1:].is_static()
    t = TimeRef()
    assert not v[:t].is_static()


def test_sum_of_dist():
    """Sanity check that running a sum on an array from a distribution
    is correctly row-wise and doesn't collapse to single summed value."""
    b = ops.Bernoulli(0.5)
    b.populate(10)
    b.sum().eval(5).shape == (10,)


def test_slice_end_t_is_correct_simpler():
    """A slice stop of none should actually technically be t+1 to be inclusive of current timestep? (This test is
    proof of why, getting index 0 shouldn't be nothing)"""
    v = Variable(Scalar(np.array([2, 3])))
    assert (v[:].eval(0) == np.array([[2], [3]])).all()
    assert v[:].eval(0).shape == np.array([[2], [3]]).shape


def test_slice_end_t_is_correct():
    """A slice stop of none should actually technically be t+1 to be inclusive of current timestep?"""
    t = TimeRef()
    m = model.Model()
    m.v0 = Variable(t + 1)
    m.v1 = Variable(Scalar(5))
    m.m0 = PostMeasurement(m.v0[m.v1 :].sum())
    ds = m()
    assert ds.m0.values[0] == 40.0


def test_piecewise_with_int():
    """Piecewise equations should support just directly specifying an integer and
    have it auto-wrapped in a scalar."""
    t = TimeRef()
    m = model.Model()
    m.v0 = Variable(Piecewise([0, 1], [t < 2, t >= 2]))
    m()


def test_interpolation():
    """Interpolation should work in regular reno math."""
    m = model.Model()
    m.v0 = Variable(Scalar([0, 1, 1.5, 2.72, 3.14]))
    m.v1 = Variable(ops.interpolate(m.v0, [1, 2, 3], [3, 2, 0]))
    ds = m(n=5, steps=1)
    np.testing.assert_almost_equal(ds.v1.values, np.array([3.0, 3.0, 2.5, 0.56, 0.0]))


def test_interpolation_pymc():
    """Interpolation should work in pymc math."""
    m = model.Model()
    m.v0 = Variable(Scalar([0, 1, 1.5, 2.72, 3.14]))
    m.v1 = Variable(ops.interpolate(m.v0, [1, 2, 3], [3, 2, 0]))
    ds = m.pymc(steps=1, compute_prior_only=True)
    np.testing.assert_almost_equal(
        ds.prior.v1.values[0][0], np.array([3.0, 3.0, 2.5, 0.56, 0.0])
    )


# TODO: add tests for slices and pymc...


# NOTE: I don't think an actual scalar on a computed value is possible, should
# always be a vector
# def test_sum_on_scalar():
#     """Running ops.sum on a scalar (likely from static varible) should give "sum", or the value
#     time the number of timesteps."""
#     v = Variable(Scalar(3))
#     v.value = 3
#     assert (ops.sum(v).eval(3) == 9)



"""Tests for individual operations and/or SD components."""

import pytest

from reno import ops, utils
from reno.components import Flow, PostMeasurement, Scalar, Stock, TimeRef, Variable
from reno.model import Model


@pytest.mark.parametrize(
    "input_list,sample_size,expected_output",
    [
        ([1, 2, 3], 2, [1, 2]),
        ([1, 2, 3], 3, [1, 2, 3]),
        ([1, 2, 3], 4, [1, 2, 3, 1]),
        ([1, 2, 3], 5, [1, 2, 3, 1, 2]),
        ([1, 2, 3], 6, [1, 2, 3, 1, 2, 3]),
    ],
)
def test_list_dist(input_list, sample_size, expected_output):
    """Specifying a list distribution and populating with various sample sizes
    should tile appropriately."""

    ldist = ops.List(input_list)
    if sample_size < len(input_list):
        with pytest.warns(RuntimeWarning):
            ldist.populate(sample_size)
    else:
        ldist.populate(sample_size)
    assert (ldist.value == expected_output).all()


def test_static_check_scalar():
    """Flows and variables whose equations are just a scalar should be static."""

    f0 = Flow(Scalar(5))
    f0.populate(5, 5)
    assert f0._static


def test_static_check_dist():
    """Flows and variables whose equations are just distributions should be static."""
    v0 = Variable(ops.List([1, 2]))
    v0.populate(5, 5)
    assert v0._static

    f0 = Flow(ops.List([1, 2]))
    f0.populate(5, 5)
    assert f0._static


def test_static_check_static_eq():
    """Flows and variables whose equations are purely static should be static."""
    v0 = Variable(Scalar(2) + 1)
    v0.populate(5, 5)
    assert v0._static

    f0 = Flow(Scalar(2) + 1)
    f0.populate(5, 5)
    assert f0._static


def test_static_check_static_eq_w_refs():
    """Flows and variables whose equations are purely static (containing refs to other
    static refs) should be static."""

    m = Model()
    m.v = Variable(Scalar(3))

    m.v0 = Variable(Scalar(2) + 1 + m.v)

    m.f0 = Flow(Scalar(2) + 1 + m.v)

    m._populate(5, 5)

    assert m.v0._static
    assert m.f0._static

    assert m.v0.is_static()
    assert m.f0.is_static()


def test_static_check_eq_w_refs():
    """Flows and variables whose equations are not static (contain non-static
    refs) should not be static."""

    t = TimeRef()

    v0 = Variable(Scalar(2) + 1 + t)
    v0.populate(5, 5)
    assert not v0._static

    f0 = Flow(Scalar(2) + 1 + t)
    f0.populate(5, 5)
    assert not f0._static


def test_static_check_eq_w_nested_refs():
    """Flows and variables whose equations are not static (contain refs that
    contain refs that are non-static, e.g. time) should not be static."""

    t = TimeRef()
    v = Variable(t)

    v0 = Variable(Scalar(2) + 1 + v)
    assert not v0.is_static()

    f0 = Flow(Scalar(2) + 1 + v)
    assert not f0.is_static()


def test_static_check_eq_w_nested_refs_to_stocks():
    """Flows and variables whose equations are not static (contain refs that
    contain refs that are non-static, e.g. a stock) should not be static."""

    s = Stock()
    v = Variable(s)

    v0 = Variable(Scalar(2) + 1 + v)
    assert not v0.is_static()

    f0 = Flow(Scalar(2) + 1 + v)
    assert not f0.is_static()


def test_static_check_static_flow_but_dynamic_limits():
    """Flows or variables whose equations are static but have non-static
    limits should not be static."""

    m = Model()
    t = TimeRef()
    m.v = Variable(Scalar(1), max=t)
    m.f = Flow(Scalar(1), max=t)

    m.v.populate(5, 5)
    m.f.populate(5, 5)

    assert not m.v._static
    assert not m.f._static

    assert not m.v.is_static()
    assert not m.f.is_static()

    m.x = Variable(Scalar(1), max=ops.minimum(Scalar(0), t))
    m.y = Variable(Scalar(1), max=ops.minimum(Scalar(0), m.f))

    assert not m.x.is_static()
    assert not m.y.is_static()


def test_static_check_eq_with_historical_value():
    """Flows or variables that reference a historical value are inherently
    based in time and can't be static."""

    v0 = Variable(Scalar(1))
    v1 = v0.history(Scalar(1))
    f0 = Flow(v1)

    v0.name = "v0"
    v1.name = "v1"
    f0.name = "f0"

    v0.populate(5, 5)
    # v1.populate(5, 5)
    f0.populate(5, 5)

    assert not v1.is_static()
    assert not f0.is_static()


def test_single_ref_eq_appears_in_seek_refs():
    """An equation that is just another reference, e.g. flow1 = var1, should
    correctly return var1 when seek_refs is called on flow1.eq"""

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.f0 = Flow(m.v0)
    # m.f0 = Flow()
    # m.f0.eq = m.v0

    assert m.f0.seek_refs() == [m.v0]
    assert m.f0.eq.seek_refs() == [m.v0]


def test_depencency_ordering_metrics():
    """Calling dependency_compute_order on a specific set of metrics should
    correctly order them."""

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.f0 = Flow(m.v0)
    m.s0 = Stock()
    m.s0 += m.f0

    m.metric1 = PostMeasurement()
    m.metric2 = PostMeasurement()

    m.metric1.eq = m.metric2 + m.f0
    m.metric2.eq = m.s0

    ordered = utils.dependency_compute_order([m.metric1, m.metric2])
    assert ordered == [m.metric2, m.metric1]


def test_submodel_getattrs():
    """Getting attributes of submodels (e.g. getattr(my_model, "submodel.attr"))
    should work."""

    m = Model(name="parent")
    s = Model(name="child")
    s.v0 = Variable(Scalar(1))
    m.s = s

    assert m.s.v0 == s.v0
    assert getattr(s, "v0") == s.v0
    assert getattr(m, "s.v0") == s.v0


def test_min_max_respected():
    """A min/max on a variable should mean that that range isn't exceeded."""
    m = Model()
    m.limited = Variable(min=0, max=5)
    m.change = Variable(6)
    m.limited.eq = m.change
    m()
    assert m.limited.value[0] == 5

    m.change.eq = -1
    m()
    assert m.limited.value[0] == 0

    m.limited.min = -2
    m()
    assert m.limited.value[0] == -1


def test_init_without_explicit_scalar():
    """Creating a stock with an initial condition that is implicitly a scalar should be auto
    converted."""
    m = Model()
    m.thing = Stock(init=5)
    m.inflow = Flow(1)
    m.thing += m.inflow
    ds = m()
    assert ds.thing.values[-1][-1] == 14.0
